/* Parameter.cpp
   implemenation of Parameter.hpp
   @author	Hyun Kim
   @date	11/20/99
*/

#include "Parameter.hpp"
#include "Population.hpp"
#include "random.hpp"
#include "utilities.hpp"
#include <iostream.h>
#include <fstream.h>
#include <string.h>

// definition of static parameters
int				Parameter::popsize;				// size of population
int				Parameter::genes;				// number of gene in chromosome
double			Parameter::riskfree;			// risk free rate
double*			Parameter::expReturn;			// expected return of each asset
double*			Parameter::stdDevReturn;		// standard deviation of return
double*			Parameter::correlation[MAXGENES];// correlation between a pair of assets
SelectionType	Parameter::selection;			// selection methods used
CrossoverType	Parameter::crossover;			// crossover operator used
double			Parameter::mutationRate;
double			Parameter::seed;				// seed for generating random number 

/* void ctor
*/
Parameter::Parameter()
{
}

/* dtor
*/
Parameter::~Parameter()
{
	delete[] Parameter::expReturn;
	delete[] Parameter::stdDevReturn;
	for ( int i = 0; i < Parameter::genes; i++ )
	{
		delete[] Parameter::correlation[ i ];
	}
}

/* read parameters from the file
*/
void readParameters( ifstream& in )
{
	char s[ 80 ] = "";
	in >> s >> Parameter::popsize;
	in >> s >> Parameter::genes;
	in >> s >> Parameter::riskfree;
	in >> s >> s;
	Parameter::selection = findSelectionOp( s );
	in >> s >> s;
	Parameter::crossover = findCrossoverOp( s );
	in >> s >> Parameter::seed;		randomize( Parameter::seed );
	in >> s;	readExpReturn( in );
	in >> s;	readStdDevReturn( in );
	in >> s;	readCorrelation( in );
	Parameter::mutationRate = 1/(double)Parameter::genes;
}

/* read the expected returns from the file
*/
void readExpReturn( ifstream& in )
{
	Parameter::expReturn = new double[ Parameter::genes ];
	for ( int i = 0; i < Parameter::genes; i++ )
	{
		in >> Parameter::expReturn[ i ];
	}
}

/* read the standard deviation of returns from the file
*/
void readStdDevReturn( ifstream& in )
{
	Parameter::stdDevReturn = new double[ Parameter::genes ];
	for ( int i = 0; i < Parameter::genes; i++ )
	{
		in >> Parameter::stdDevReturn[ i ];
	}
}

/* read correlation from the file
*/
void readCorrelation( ifstream& in )
{
	int i, j;
	for ( i = 0; i < Parameter::genes; i++ )
	{
		Parameter::correlation[ i ] = new double[ Parameter::genes ];
	}
	for ( i = 0; i < Parameter::genes; i++ )
	{
		for ( j = i; j < Parameter::genes; j++ )
		{
			in >> Parameter::correlation[ i ][ j ];
			Parameter::correlation[ j ][ i ] = Parameter::correlation[ i ][ j ];
		}
	}
}

/* return the appropriate type of the selection operator
   @param s string of selection method
   @return SelectionType
*/
SelectionType findSelectionOp( char* s )
{
	if ( !strcmp( s, "roulette_wheel_selection" ) )
	{
		return roulette_wheel_selection;
	}
	else if ( !strcmp( s, "tournament_with_replacement" ) )
	{
		return tournament_with_replacement;
	}
	else
	{
		// default selection operator
		return roulette_wheel_selection;
	}
}

/* return the appropriate type of the crossover operator
   @param s string of crossover operator
   @return CrossoverType
*/
CrossoverType findCrossoverOp( char* s )
{
	if ( !strcmp( s, "ptx_1" ) )
	{
		return ptx_1;
	}
	else if ( !strcmp( s, "ptx_2" ) )
	{
		return ptx_2;
	}
	else if ( !strcmp( s, "uniform" ) )
	{
		return uniform;
	}
	else if ( !strcmp( s, "linx" ) )
	{
		return linx;
	}
	else if ( !strcmp( s, "alinx_1" ) )
	{
		return alinx_1;
	}
	else if ( !strcmp( s, "alinx_2" ) )
	{
		return alinx_2;
	}
	else if ( !strcmp( s, "alinx_3" ) )
	{
		return alinx_3;
	}
	else if ( !strcmp( s, "alinx_4" ) )
	{
		return alinx_4;
	}
	else
	{
		// default crossover operator
		return ptx_1;
	}
}

/* print the parameters to the file
   @param out file output stream
*/
void printParameters( ofstream& out )
{
	int i, j;
	out << "GA settings:" << endl << endl
		<< "------------------------------------------------------------" << endl
		<< "population size       : " << Parameter::popsize << endl
		<< "number of genes       : " << Parameter::genes << endl
		<< "risk-free rate        : " << Parameter::riskfree << endl
		<< "selection scheme      : " << Parameter::selection << endl
		<< "crossover operator    : " << Parameter::crossover << endl
		<< "mutation rate         : " << Parameter::mutationRate << endl
		<< "random seed           : " << Parameter::seed << endl
		<< "expected return       : ";
	for ( i = 0; i < Parameter::genes; i++ )
	{
		out << Parameter::expReturn[ i ] << " ";
	}
	out << endl;
	out << "Std. Dev. of Return   : ";
	for ( i = 0; i < Parameter::genes; i++ )
	{
		out << Parameter::stdDevReturn[ i ] << " ";
	}
	out << endl;
	out << "correlation           : ";
	for ( i = 0; i < Parameter::genes; i++ )
	{
		for ( j = i; j < Parameter::genes; j++ )
		{
			out << Parameter::correlation[ i ][ j ] << " ";
		}
	}
	out << endl << endl
		<< "------------------------------------------------------------"
		<< endl;
}

/* print the population matrix for a certain generation
   @param f file output stream
   @param gen generation number
   @param pop pointer to the population
*/
void report( ofstream& f, int gen, Population* pop )
{
	f << "Generation                      : " << gen << endl
	  << "Fitness (max, avg, min)         : "
	  << pop->getMaxFitness() << " " << pop->getAvgFitness() << " " << pop->getMinFitness() << endl
	  << endl;
	f << "Population Matrix (with fitness): " << endl 
	  << endl;
	for ( int i = 0; i < Parameter::popsize; i++ )
	{
		f << (*pop)[ i ] << ": " << (*pop)[ i ].getFitness() << endl;
	}
	f << endl
	  << "------------------------------------------------------------" 
	  << endl;
}

/* print the information for analysis
   @param f file output stream
   @param gen generation number
   @param pop pointer to the population
*/
void analyze( ofstream& f, int gen, Population* pop )
{
	f << pop->getMaxFitness() << endl;
}

/* print the result
   @param f file output stream
   @param pop pointer to the population
*/
void result( ofstream& f, Population* pop )
{
	f << "Expected Return                 : " << getExpReturn( (*pop)[ pop->getMaxIndex() ] )
	  << endl
	  << "Std. Dev. ( Risk )              : " << getStdDev( (*pop)[ pop->getMaxIndex() ] )
	  << endl
	  << "Max Fitness                     : " << pop->getMaxFitness()
	  << endl
	  << "Soultion                        : " << (*pop)[ pop->getMaxIndex() ] 
	  << endl
	  << endl
	  << "------------------------------------------------------------" 
	  << endl;
}
