/* Population.cpp
   implementaion of Population.hpp
   @author	Hyun Kim
   @date	11/20/99
*/

#include "Population.hpp"
#include "Parameter.hpp"
#include "random.hpp"
#include "utilities.hpp"
#include <assert.h>

/* void ctor
*/
Population::Population()
{
	chromosomes = new Chromosome[ Parameter::popsize ];
	assert( chromosomes != 0 );
	setSize( Parameter::popsize );
}

/* copy ctor
*/
Population::Population( Population& pop )
{
	chromosomes = new Chromosome[ Parameter::popsize ];
	assert( chromosomes != 0 );
	for ( int i = 0; i < Parameter::popsize; i++ )
	{
		chromosomes[ i ] = pop[ i ];
	}
	size = pop.getSize();
	sumFitness = pop.getSumFitness();
	maxFitness = pop.getMaxFitness();
	minFitness = pop.getMinFitness();
	avgFitness = pop.getAvgFitness();
	maxIndex = pop.getMaxIndex();
	minIndex = pop.getMinIndex();
	child1 = pop.getChild1();
	child2 = pop.getChild2();
}

/* dtor
*/
Population::~Population()
{
	delete[] chromosomes;
}


/* return the i-th chromosome in the population
   @param i i-th chromosome in the population
   @return Chromosome
*/
Chromosome& Population::operator[]( int i )
{
	assert( i >= 0 && i < size );
	return chromosomes[ i ];
}

/* assign the population to another population
   @param pop current population
   @return Population
*/
Population& Population::operator=( Population& pop )
{
	if( this == &pop ) 
	{
		return *this;
	}

	delete[] chromosomes;
	chromosomes = new Chromosome[ Parameter::popsize ];
	assert( chromosomes != 0 );
	size = pop.getSize();
	for ( int i = 0; i < pop.getSize(); i++ )
	{
		chromosomes[ i ] = pop[ i ];
	}
	sumFitness = pop.getSumFitness();
	maxFitness = pop.getMaxFitness();
	minFitness = pop.getMinFitness();
	avgFitness = pop.getAvgFitness();
	maxIndex = pop.getMaxIndex();
	minIndex = pop.getMinIndex();
	child1 = pop.getChild1();
	child2 = pop.getChild2();
	return *this;
}

/* initialize the population
*/
void Population::initialize()
{
	for ( int i = 0; i < size; i++ )
	{
		chromosomes[ i ].random();
		chromosomes[ i ].evaluate();
	}
	statistics();
}

/* select the any chromosome in the population for mating
   @return int
*/
int Population::selection()
{
	int i = -1;		// chosen chromosome
	switch( Parameter::selection )
	{
	case roulette_wheel_selection:
		i = rouletteWheelSelection();
		break;
	case tournament_with_replacement:
		break;
	default:
		break;
	}
	return i;
}

/* apply the crossover operator to parents
   @param parent1 one parent
   @param parent2 another parent
*/
void Population::crossover( Chromosome& parent1, Chromosome& parent2 )
{
	switch( Parameter::crossover )
	{
	case ptx_1:
		doPTX1( parent1, parent2 );
		break;
	case ptx_2:
		break;
	case uniform:
		doUNIFORM( parent1, parent2 );
		break;
	case linx:
		doLINX( parent1, parent2 );
		break;
	case alinx_1:
		break;
	case alinx_2:
		break;
	case alinx_3:
		break;
	case alinx_4:
		break;
	default:
		break;
	}
/*	child1.mutate();	// mutation does not include adjusting function
	child2.mutate();
*/
	child1.adjust( child1.getLength() );
	child2.adjust( child2.getLength() );

	// after adjusting, apply transpose
	child1.transpose();
	child2.transpose();

	child1.evaluate();
	child2.evaluate();
}

/* reproduce the population from the current popuation by selection and crossover
   @return Population
*/
Population Population::reproduce()
{
	int i = 0;		// number of generations
	Population newPop;

	while ( i < size )
	{
		int p1 = selection();
		int p2 = selection();
		if ( p1 != p2 )
		{
			// produce offsprings
			crossover( chromosomes[ p1 ], chromosomes[ p2 ] );
			newPop[ i++ ] = child1;
			newPop[ i++ ] = child2;
			newPop.statistics();
		}
		else
		{
			newPop = reproduce();
		}
	}
	return newPop;
}

/* reproduce the population from the current population. 
   It replaces the poor solution by offsprings.
   @return Population
*/
Population Population::reproduceWithReplacement()
{
	Population newPop;
	int p1 = selection();
	int p2 = selection();
	if ( p1 != p2 )
	{
		// produce offsprings
		crossover( chromosomes[ p1 ], chromosomes[ p2 ] );
		newPop = *this;
		int minIndex1 = getMinIndex();
		int minIndex2 = getIndexAt( Parameter::popsize - 1 );
		newPop[ minIndex1 ] = child1;
		newPop[ minIndex2 ] = child2;
		newPop.statistics();
	}
	else
	{
		newPop = reproduceWithReplacement();
	}
	return newPop;
}

/* compute the total/max/avg/min fitness of the population
*/
void Population::statistics()
{
	// initialize
	int i = 0;
	sumFitness = chromosomes[ i ].getFitness();
	maxFitness = chromosomes[ i ].getFitness();
	minFitness = chromosomes[ i ].getFitness();
	avgFitness = chromosomes[ i ].getFitness();
	maxIndex = i;
	minIndex = i;

	// find the total/max/avg/min fitness
	for ( i = 1; i < size; i++ )
	{
		double newFitness = chromosomes[ i ].getFitness();
		sumFitness += newFitness;
		if ( newFitness > maxFitness )
		{
			maxFitness = newFitness;
			maxIndex = i;
		}
		else if ( newFitness < minFitness )
		{
			minFitness = newFitness;
			minIndex = i;
		}
	}
	avgFitness = sumFitness / size;
}

/* select any chromosome in the population using roulette wheel selection
   @return int
*/
int Population::rouletteWheelSelection()
{
	double rand = getSumFitness() * random01();	// random point on wheel
	double partialSum = 0;
	int i = -1;
	while ( partialSum < rand && i < size )
	{
		i++;
		partialSum += chromosomes[ i ].getFitness();
	} 

	return i;
}

/* apply the 1 point crossover operator
   @param parent1 one parent for crossover
   @param parent2 another parent for crossover
*/
void Population::doPTX1( Chromosome& parent1, Chromosome& parent2 )
{
	int length = parent1.getLength();
	int xPoint = rnd( 0, length-1 );	// crossover point
	int i;
	for ( i = 0; i < xPoint; i++ )
	{
		child1[ i ] = parent1[ i ];
		child2[ i ] = parent2[ i ];
	}
	for ( i = xPoint; i < length; i++ )
	{
		child1[ i ] = parent2[ i ];
		child2[ i ] = parent1[ i ];
	}
}

/* apply the uniform crossover operator
   @param parent1 one parent for crossover
   @param parent2 another parent for crossover
*/
void Population::doUNIFORM( Chromosome& parent1, Chromosome& parent2 )
{
	double r;
	for( int i = 0; i < Parameter::genes; i++ )
	{
		r = random01();
		if( r < 0.5 )
		{
			child1[ i ] = parent1[ i ];
			child2[ i ] = parent2[ i ];
		}
		else
		{
			child1[ i ] = parent2[ i ];
			child2[ i ] = parent1[ i ];
		}
	}
}

/* apply the LinX crossover operator
   @param parent1 one parent for crossover
   @param parent2 another parent for crossover
*/
void Population::doLINX( Chromosome& parent1, Chromosome& parent2 )
{
	child1 = LINXchild( parent1, parent2 );
	child2 = LINXchild( parent1, parent2 );
}

/* generate a chile from parents using LinX crossover operator
   @param parent1 one parent for crossover
   @param parent2 another parent for crossover
   @return Chromosome
*/
Chromosome Population::LINXchild( Chromosome& parent1, Chromosome& parent2 )
{
	Chromosome child;
	int length = parent1.getLength();
	// s1(or s2) is the set of offspring's genes inherited so far from parent1(or parent2)
	int* s1 = new int[ Parameter::genes ];
	int* s2 = new int[ Parameter::genes ];	
	int cntS1 = 0, cntS2 = 0;	// number of s1(or s2)
	
	int i, j, visited[12];
	double linkage[12][12];

	// compute linkage_matrix
	// Note: linkage_matrix = (1 - correlation_matrix) / 2, since -1 <= correlation_matrix <= 1.
	for ( i = 0; i < 12; i++ )
	{
		visited[i] = 0;		// initialize the visited array
		for ( j = 0; j < 12; j++ )
		{
			linkage[i][j] = (1 - Parameter::correlation[i][j])/2;
		}
	}

	// For two randomly chosen genes, select one allele from parent1 and the other from parent2
	int point1 = rnd( 0, length - 1 ), point2 = rnd( 0, length - 1 );
	if ( point1 == point2 )	
		point2 = ( point2 + 1 ) % length;
	child[ point1 ] = parent1[ point1 ];
	s1[ cntS1++ ] = point1;
	child[ point2 ] = parent2[ point2 ];
	s2[ cntS2++ ] = point2;
	visited[ point2 ] = visited[ point2 ] = 1;

	// copy all common genes from parents to offspring
	for ( i = 0; i < length; i++ )
	{
		if ( parent1[i] == parent2[i] )
		{
			child[i] = parent1[i];
			visited[i] = 1;
		}
	}
	for ( i = 0; i < length; i++ )
	{
		if ( visited[i] )
			continue;

		// compute h1 and h2
		double h1 = 1, h2 = 1;
		for ( j = 0; j < cntS1; j++ )
		{
			h1 = h1 * linkage[ s1[j] ][ i ];
			h2 = h2 * ( 1 - linkage[ s1[j] ][ i ] );
		}
		for ( j = 0; j < cntS2; j++ )
		{
			h1 = h1 * ( 1 - linkage[ s2[j] ][ i ] );
			h2 = h2 * linkage[ s2[j] ][ i ];
		}

		double random = random01();		// generate any random number in [0, 1]
		if ( random <= h1 / ( h1 + h2 ) )
		{
			child[ i ] = parent1[ i ];
			s1[ cntS1++ ] = i;
		}
		else
		{
			child[ i ] = parent2[ i ];
			s2[ cntS2++ ] = i;
		}
		visited[i] = 1;
	}
	return child;
}

/* return the index which indicates the chromosome with the N-th largest fitness
   @param n n-th largest fitness
   @return int
*/
int Population::getIndexAt( int n )
{
	int* sorter = new int[ Parameter::popsize ];

	for ( int i = 0; i < Parameter::popsize; i++ )
	{
		sorter[ i ] = i;
	}
	sort( *this, sorter, Parameter::popsize );

	return sorter[ n - 1 ];
}